



python -c "from numpy.core._multiarray_umath import __cpu_features__; print(f'Testing version with AVX2-support - ' + str(__cpu_features__['AVX2']))"
IF %ERRORLEVEL% NEQ 0 exit /B 1
pytest tests --log-file-level=INFO --log-file=log.txt -k "not TestComputeGT"
IF %ERRORLEVEL% NEQ 0 exit /B 1
type log.txt
IF %ERRORLEVEL% NEQ 0 exit /B 1
python -c "q = open('log.txt').readlines(); import sys; sys.exit(0 if 'Successfully loaded faiss with AVX2 support.' in [x[35:-1] for x in q] else 1)"
IF %ERRORLEVEL% NEQ 0 exit /B 1
set NPY_DISABLE_CPU_FEATURES=AVX2
IF %ERRORLEVEL% NEQ 0 exit /B 1
python -c "from numpy.core._multiarray_umath import __cpu_features__; print(f'Testing version with AVX2-support - ' + str(__cpu_features__['AVX2']))"
IF %ERRORLEVEL% NEQ 0 exit /B 1
pytest tests --log-file-level=INFO --log-file=log.txt -k "not TestComputeGT"
IF %ERRORLEVEL% NEQ 0 exit /B 1
type log.txt
IF %ERRORLEVEL% NEQ 0 exit /B 1
python -c "q = open('log.txt').readlines(); import sys; sys.exit(0 if 'Successfully loaded faiss.' in [x[35:-1] for x in q] else 1)"
IF %ERRORLEVEL% NEQ 0 exit /B 1
exit /B 0
